package Irpg::Irc;
=head1 NAME
Irpg/Irc.pm - Functions to interact with a irc server

=head1 SYNOPSIS
    use Irpg::Irc;

=head1 DESCRIPTION
    This module manage the connection to IRC and provides 
    functions to interact with this connection.

=head2 Exports
=over
=item
=back
=cut

use strict;
use warnings;
use IO::Socket;
use IO::Select;
use Irpg::Utils;
use Exporter qw(import);
our @EXPORT = qw($sock);
our @EXPORT_OK = qw($sock @queue
                    &irc_connect &irc_clean
                    &checkmsg &fq &sts
                    &privmsg &chanmsg &notice $inbytes $outbytes);
our %EXPORT_TAGS = (connection=>[qw($sock &irc_connect &irc_clean &checkmsg)],
                    interaction=>[qw(fq sts privmsg chanmsg notice)]);


my $outbytes = 0; # sent bytes
my $inbytes = 0; # received bytes

my @queue; # outgoing message queue

my $sel; # IO::Select object
my $buffer; # buffer for socket stuff
my $conn_tries = 0; # number of connection tries. gives up after trying each
                    # server twice
my $sock; # IO::Socket::INET object
my $sock_ref = \$sock;
my $freemessages = 4; # number of "free" privmsgs we can send. 0..$freemessages

my $opts;

my $silentmode_ref; 
=head1 FUNCTION init_pkg
    This function sets the references to
        options and players hashes.
=over
=item SCALAR (ref)  - reference to the options hash
=item SCALAR (ref)  - reference to the players hash
=cut
sub init_pkg { $silentmode_ref = shift;}


=head1 FUNCTION irc_connect
    This function manage the connection to the irc server.
=over
=item SCALAR (ref)    - reference to the options hash
=back
=cut
sub irc_connect {
    $opts = shift;
    while (!$sock && $conn_tries < 2*@{$opts->{servers}}) {
        debug("Connecting to $opts->{servers}->[0]...");
        my %sockinfo = (PeerAddr => $opts->{servers}->[0],
                        PeerPort => 6667);
        if ($opts->{localaddr}) { $sockinfo{LocalAddr} = $opts->{localaddr}; }
        $sock = IO::Socket::INET->new(%sockinfo) or
            debug("Error: failed to connect: $!\n");
        ++$conn_tries;
        if (!$sock) {
            # cycle front server to back if connection failed
            push(@{$opts->{servers}},shift(@{$opts->{servers}}));
        }
        else { debug("Connected."); }
    }

    if (!$sock) {
        debug("Error: Too many connection failures, exhausted server list.\n",1);
    }

    $conn_tries=0;

    $sel = IO::Select->new($sock);

    sts("NICK $opts->{botnick}");
    sts("USER $opts->{botuser} 0 0 :$opts->{botrlnm}");
}

=head1 FUNCTION irc_clean
    This function frees the ressources used for the connection.
=over
=item SCALAR (ref)    - reference to the options hash
=back
=cut
sub irc_clean {
    my $opts = shift;
    undef(@queue);
    undef($sock);
    debug("Socket closed; disconnected. Cleared outgoing message ".
          "queue. Waiting $opts->{reconnect_wait}s before next ".
          "connection attempt...");
    sleep($opts->{reconnect_wait});
}


=head1 FUNCTION checkmsg
    This function checks incoming messages, and returns
    a list of 2 elements:
    one boolean set to 1 if all went fine and 0 if we are disconnected
    and the string of the message (may be undef)
=cut
sub checkmsg {
    my($readable) = IO::Select->select($sel,undef,undef,0.5);
    my @ret = ();
    if (defined($readable)) {
        my $fh = $readable->[0];
        my $buffer2;
        $fh->recv($buffer2,512,0);
        if (length($buffer2)) {
            push @ret, 1;
            $buffer .= $buffer2;
            while (index($buffer,"\n") != -1) {
                my $line = substr($buffer,0,index($buffer,"\n")+1);
                $buffer = substr($buffer,length($line));
                push @ret, $line;
            }
        }
        else {
            # uh oh, we've been disconnected from the server
            push @ret, 0;

            close($fh);
            $sel->remove($fh);

        }
    }
    else { select(undef,undef,undef,1); }
    return @ret ? @ret : (1);
}

=head1 FUNCTION fq
    deliver message(s) from queue
=over
=item SCALAR (ref)    - reference to the options hash
=back
=cut
sub fq { # deliver message(s) from queue
    my $opts = shift;
    if (!@queue) {
        ++$freemessages if $freemessages < 4;
        return undef;
    }
    my $sentbytes = 0;
    for (0..$freemessages) {
        last() if !@queue; # no messages left to send
        # lower number of "free" messages we have left
        my $line=shift(@queue);
        # if we have already sent one message, and the next message to be sent
        # plus the previous messages we have sent this call to fq() > 768 bytes,
        # then requeue this message and return. we don't want to flood off,
        # after all
        if ($_ != 0 && (length($line)+$sentbytes) > 768) {
            unshift(@queue,$line);
            last();
        }
        if ($sock) {
            debug("(fm$freemessages) -> $line");
            --$freemessages if $freemessages > 0;
            print $sock "$line\r\n";
            $sentbytes += length($line) + 2;
        }
        else {
            undef(@queue);
            debug("Disconnected: cleared outgoing message queue.");
            last();
        }
        $outbytes += length($line) + 2;
    }
}


=head1 FUNCTION sts
    Sends data to server
=over
=item SCALAR (ref)    - reference to the options hash
=back
=cut
sub sts { # send to server
    my($text,$skipq) = @_;
    if ($skipq) {
        if ($sock) {
            print $sock "$text\r\n";
            $outbytes += length($text) + 2;
            debug("-> $text");
        }
        else {
            # something is wrong. the socket is closed. clear the queue
            undef(@queue);
            debug("\$sock isn't writeable in sts(), cleared outgoing queue.\n");
            return;
        }
    }
    else {
        push(@queue,$text);
        debug(sprintf("(q%03d) = %s\n",$#queue,$text));
    }
}

=head1 FUNCTION privmsg
    Sends a message to an arbitrary entity
=over
=item SCALAR (string)    - message to send
=item SCALAR (string)    - target
=back
=cut
sub privmsg { # send a message to an arbitrary entity
    my $msg = shift or return undef;
    my $target = shift or return undef;
    my $force = shift;
    if (($$silentmode_ref == 3 || ($target !~ /^[\+\&\#]/ && $$silentmode_ref == 2))
        && !$force) {
        return undef;
    }
    while (length($msg)) {
        sts("PRIVMSG $target :".substr($msg,0,450),$force);
        substr($msg,0,450)="";
    }
}

=head1 FUNCTION pubmsg
    Sends a message to the channel
=over
=item SCALAR (string)    - message to send
=back
=cut
sub chanmsg { # send a message to the channel
    my $msg = shift or return undef;
    if ($$silentmode_ref & 1) { return undef; }
    privmsg($msg, $opts->{botchan}, shift);
}

=head1 FUNCTION notice
    Sends a notice to an arbitrary entity
=over
=item SCALAR (string)    - message to send
=item SCALAR (string)    - target
=item SCALAR (bool)      - force (default false)
=back
=cut
sub notice { # send a notice to an arbitrary entity
    my $msg = shift or return undef;
    my $target = shift or return undef;
    my $force = shift;
    if (($$silentmode_ref == 3 || ($target !~ /^[\+\&\#]/ && $$silentmode_ref == 2))
        && !$force) {
        return undef;
    }
    while (length($msg)) {
        sts("NOTICE $target :".substr($msg,0,450),$force);
        substr($msg,0,450)="";
    }
}

1;
